"""AUTOGENERATED. DO NOT EDIT."""

from keras.api._v1.keras.datasets import boston_housing
from keras.api._v1.keras.datasets import cifar10
from keras.api._v1.keras.datasets import cifar100
from keras.api._v1.keras.datasets import fashion_mnist
from keras.api._v1.keras.datasets import imdb
from keras.api._v1.keras.datasets import mnist
from keras.api._v1.keras.datasets import reuters
